
/*
 *    ChunkSync - create space-efficient incremental (remote) backups of large files
 *    Copyright (C) 2009  Florian Zumbiehl <florz@florz.de>
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 *
 *    In addition, as a special exception, the author of this
 *    program gives permission to link the code of its 
 *    release with the OpenSSL project's "OpenSSL" library (or
 *    with modified versions of it that use the same license as
 *    the "OpenSSL" library), and distribute the linked
 *    executables. You must obey the GNU General Public 
 *    License in all respects for all of the code used other 
 *    than "OpenSSL".  If you modify this file, you may extend
 *    this exception to your version of the file, but you are
 *    not obligated to do so.  If you do not wish to do so,
 *    delete this exception statement from your version.
 */

#include <stdbool.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <pthread.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/wait.h>

#include "defs.h"
#include "io.h"
#include "err.h"
#include "sender.h"
#include "receiver.h"
#include "progress.h"
#include "frontend.h"

struct xfh sender_fh,receiver_fh;

static pid_t sender_pid,receiver_pid;
static pthread_t sender_stderr_forwarder,receiver_stderr_forwarder,sender_forwarder,receiver_forwarder;
static pthread_mutex_t forwarders_lock=PTHREAD_MUTEX_INITIALIZER;
static bool forwarders_launched,forwarders_exiting;
static void *forwarders_error;
static pthread_cond_t forwarders_exiting_cond=PTHREAD_COND_INITIALIZER;
static pthread_t forwarders_exiting_thr;

struct stderr_forwarder_params {
	const char *child_role;
	int fd;
};

static void pipe2(int *w,int *r){
	int fds[2],flags;

	if(pipe(fds))die(true,"pipe()");
	for(int x=2;x--;){
		if((flags=fcntl(fds[x],F_GETFD))<0)die(true,"fcntl(pipe,F_GETFD)");
		if(fcntl(fds[x],F_SETFD,flags|FD_CLOEXEC))die(true,"fcntl(pipe,F_SETFD)");
	}
	*r=fds[0];
	*w=fds[1];
}

static void *stderr_forward(void *params_arg){
	struct stderr_forwarder_params *params=params_arg;
	size_t pos=0;
	ssize_t r;
	char buf[64*1024];

	while(true){
		if((r=read(params->fd,buf+pos++,1))!=1){
			if(r)die(true,"read(%s_stderr)",params->child_role);
			if(close(params->fd))die(true,"close(%s_stderr)",params->child_role);
			free(params);
			pthread_exit(NULL);
		}
		if(pos==sizeof(buf)-1)buf[pos++]='\n';
		if(buf[pos-1]=='\n'){
			err_start(false);
			err_printname(params->child_role);
			if(fwrite(buf,1,pos,stderr)!=pos)die(true,"fwrite(stderr)");
			err_end();
			pos=0;
		}
	}
}

static void __attribute__ ((noreturn)) final_cleanup(int exitcode){
	pid_t pid;

	xfh_close(&sender_fh);
	if(receiver_pid)xfh_close(&receiver_fh);
	if((errno=pthread_join(sender_stderr_forwarder,NULL)))die(true,"pthread_join()");
	if(receiver_pid&&(errno=pthread_join(receiver_stderr_forwarder,NULL)))die(true,"pthread_join()");
	while(sender_pid||receiver_pid){
		if((pid=wait(NULL))<0)die(true,"wait()");
		if(pid==sender_pid)sender_pid=0;
		if(pid==receiver_pid)receiver_pid=0;
	}
	exit(exitcode);
}

static void frontend_abort(bool error){
	bool my_forwarders_launched;
	void *my_forwarders_error;

	pthread_mutex_lock(&forwarders_lock);
	my_forwarders_launched=forwarders_launched;
	my_forwarders_error=forwarders_error;
	pthread_mutex_unlock(&forwarders_lock);
	if(error){
		if(my_forwarders_launched)pthread_exit(PTHREAD_CANCELED);
			else final_cleanup(1);
	}else if(my_forwarders_error)pthread_exit(PTHREAD_CANCELED);
}

static void forwarder_exiting(void *ignored){
	pthread_mutex_lock(&forwarders_lock);
	forwarders_exiting=true;
	forwarders_exiting_thr=pthread_self();
	pthread_cond_signal(&forwarders_exiting_cond);
	pthread_mutex_unlock(&forwarders_lock);
}

static void *forwarder_launch(void *func){
	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL);
	pthread_cleanup_push(&forwarder_exiting,NULL);
	((void (*)(void))func)();
	pthread_exit(NULL);
	pthread_cleanup_pop(0);
}

static void backend_launch(pid_t *backend_pid,struct xfh *frontend_fh,pthread_t *stderr_forwarder_thr,char *child_role,char *arg){
	int frontend_fds[STDERR_FILENO+1],backend_fds[STDERR_FILENO+1],argc;
	struct stderr_forwarder_params *stderr_forwarder_params;
	char buf[strlen(PROTOCOL_MAGIC)],*host,*port,*path,*argv[10];

	if(!(arg=strdup(arg)))die(true,"strdup()");
	port=NULL;
	if(*arg=='['){
		host=arg+1;
		if(!(path=strchr(host,']'))||path[1]!=':')die(false,"invalid host/path syntax");
		*path=0;
		path+=2;
	}else if((path=strchr(arg,':'))){
		host=arg;
		*(path++)=0;
	}else{
		host="";
		path=arg;
	}
	if(!*host)host=NULL;
		else if((port=strrchr(host,'/')))*(port++)=0;
	pipe2(frontend_fds+STDIN_FILENO,backend_fds+STDIN_FILENO);
	pipe2(backend_fds+STDOUT_FILENO,frontend_fds+STDOUT_FILENO);
	pipe2(backend_fds+STDERR_FILENO,frontend_fds+STDERR_FILENO);
	if(!(stderr_forwarder_params=malloc(sizeof(*stderr_forwarder_params))))die(true,"malloc()");
	stderr_forwarder_params->child_role=child_role;
	stderr_forwarder_params->fd=frontend_fds[STDERR_FILENO];
	if((errno=pthread_create(stderr_forwarder_thr,NULL,stderr_forward,stderr_forwarder_params)))die(true,"pthread_create()");
	xfh_init(frontend_fh,frontend_fds[STDOUT_FILENO],frontend_fds[STDIN_FILENO],&frontend_abort,child_role);
	if(!(*backend_pid=fork())){
		err_set_backend_mode(true);
		for(int x=STDERR_FILENO+1;x--;)if(dup2(backend_fds[x],x)<0)die(true,"dup2()");
		if(signal(SIGPIPE,SIG_DFL)==SIG_ERR)die(true,"signal()");
		argc=0;
		if(host){
			argv[argc++]="ssh";
			argv[argc++]="-T";
			if(port){
				argv[argc++]="-p";
				argv[argc++]=port;
			}
			argv[argc++]="--";
			argv[argc++]=host;
		}
		argv[argc++]="chunksync";
		argv[argc++]="-b";
		argv[argc++]=child_role;
		argv[argc]=NULL;
		execvp(*argv,argv);
		die(true,"exec(%s)",child_role);
	}
	if(*backend_pid<0)die(true,"fork(%s)",child_role);
	for(int x=STDERR_FILENO+1;x--;)if(close(backend_fds[x]))die(true,"close()");
	xread(frontend_fh,buf,strlen(PROTOCOL_MAGIC));
	if(memcmp(buf,PROTOCOL_MAGIC,strlen(PROTOCOL_MAGIC)))die(false,"incorrect protocol magic from %s",child_role);
	xwrite(frontend_fh,PROTOCOL_MAGIC,strlen(PROTOCOL_MAGIC));
	if(xread_u8(frontend_fh))die(false,"incompatible protocol version from %s",child_role);
	xread_u8(frontend_fh);
	xwrite_u8(frontend_fh,0);
	xwrite_u8(frontend_fh,0);
	xwrite_u8(frontend_fh,0);
	xwrite_string(frontend_fh,path);
	free(arg);
}

void frontend_main(bool quiet,size_t linkdirs_cnt,char **linkdirs_names,int argc,char **argv){
	off_t chunk_size;
	char *end;
	pthread_t second_thread;
	void *second_return;

	if(argc!=3)usage(true);
	errno=0;
	chunk_size=strtoll(*argv,&end,10);
	if(errno||*end||chunk_size<1)die(false,"invalid chunk size");
	progress_set_quiet(quiet);
	backend_launch(&sender_pid,&sender_fh,&sender_stderr_forwarder,"sender",argv[1]);
	if(xread_u8(&sender_fh)!=1)proterr();
	xwrite_u8(&sender_fh,linkdirs_cnt);
	xwrite_u64(&sender_fh,chunk_size);
	backend_launch(&receiver_pid,&receiver_fh,&receiver_stderr_forwarder,"receiver",argv[2]);
	if(xread_u8(&receiver_fh)!=2)proterr();
	xwrite_u8(&receiver_fh,linkdirs_cnt);
	for(size_t x=0;x<linkdirs_cnt;x++)xwrite_string(&receiver_fh,linkdirs_names[x]);
	pthread_mutex_lock(&forwarders_lock);
	forwarders_launched=true;
	pthread_mutex_unlock(&forwarders_lock);
	if((errno=pthread_create(&sender_forwarder,NULL,forwarder_launch,sender_forward)))die(true,"pthread_create()");
	if((errno=pthread_create(&receiver_forwarder,NULL,forwarder_launch,receiver_forward)))die(true,"pthread_create()");
	pthread_mutex_lock(&forwarders_lock);
	while(!forwarders_exiting)pthread_cond_wait(&forwarders_exiting_cond,&forwarders_lock);
	second_thread=pthread_equal(forwarders_exiting_thr,sender_forwarder)?receiver_forwarder:sender_forwarder;
	if((errno=pthread_join(forwarders_exiting_thr,&forwarders_error)))die(true,"pthread_join()");
	if(forwarders_error)pthread_cancel(second_thread);
	pthread_mutex_unlock(&forwarders_lock);
	if((errno=pthread_join(second_thread,&second_return)))die(true,"pthread_join()");
	pthread_mutex_lock(&forwarders_lock);
	final_cleanup(forwarders_error||second_return);
}

