
/*
 *    ChunkSync - create space-efficient incremental (remote) backups of large files
 *    Copyright (C) 2009  Florian Zumbiehl <florz@florz.de>
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 *
 *    In addition, as a special exception, the author of this
 *    program gives permission to link the code of its 
 *    release with the OpenSSL project's "OpenSSL" library (or
 *    with modified versions of it that use the same license as
 *    the "OpenSSL" library), and distribute the linked
 *    executables. You must obey the GNU General Public 
 *    License in all respects for all of the code used other 
 *    than "OpenSSL".  If you modify this file, you may extend
 *    this exception to your version of the file, but you are
 *    not obligated to do so.  If you do not wish to do so,
 *    delete this exception statement from your version.
 */

#include <stdbool.h>
#include <pthread.h>
#include <errno.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>

#include "utils.h"
#include "err.h"
#include "progress.h"

static pthread_t thr;
static pthread_mutex_t lock=PTHREAD_MUTEX_INITIALIZER;
static pthread_cond_t cond=PTHREAD_COND_INITIALIZER;
static bool quiet,visible,redraw,do_exit;
static off_t chunks_total,chunks_completed,chunks_transmitted;

static void *printer(void *ignored){
	off_t my_chunks_completed=0,my_chunks_transmitted=0;
	bool my_do_exit=false;

	while(true){
		err_start(true);
		if(fprintf(stderr,"\rcompleted %lli/%lli (%i%%), transmitted %lli (%i%%)\e[K",
			my_chunks_completed,chunks_total,chunks_total?(int)(((float)my_chunks_completed)/chunks_total*100):100,
			my_chunks_transmitted,my_chunks_completed?(int)(((float)my_chunks_transmitted)/my_chunks_completed*100):0)<0)
				die(true,"fprintf(stderr)");
		visible=true;
		if(my_do_exit)progress_clear();
		err_end();
		if(my_do_exit)pthread_exit(NULL);
		usleep(50000);
		pthread_mutex_lock(&lock);
		while(!redraw)
			if((errno=pthread_cond_wait(&cond,&lock)))die(true,"pthread_cond_wait()");
		redraw=false;
		my_chunks_completed=chunks_completed;
		my_chunks_transmitted=chunks_transmitted;
		my_do_exit=do_exit;
		pthread_mutex_unlock(&lock);
	}
}

static void trigger_redraw(void){
	if(!redraw){
		redraw=true;
		if((errno=pthread_cond_signal(&cond)))die(true,"pthread_cond_signal()");
	}
}

void progress_clear(){
	if(visible){
		if(fputc('\n',stderr)==EOF)die(true,"fputc(stderr)");
		visible=false;
	}
}

void progress_set_quiet(bool quiet_param){
	quiet=quiet_param;
}

void progress_update(bool transmitted){
	if(quiet)return;
	pthread_mutex_lock(&lock);
	if(chunks_completed==ANYSINT_MAX(off_t))proterr();
	chunks_completed++;
	if(transmitted)chunks_transmitted++;
	trigger_redraw();
	pthread_mutex_unlock(&lock);
}

void progress_launch(off_t chunks){
	if(quiet)return;
	chunks_total=chunks;
	if((errno=pthread_create(&thr,NULL,printer,NULL)))die(true,"pthread_create()");
}

void progress_stop(){
	if(quiet)return;
	pthread_mutex_lock(&lock);
	do_exit=true;
	trigger_redraw();
	pthread_mutex_unlock(&lock);
	if((errno=pthread_join(thr,NULL)))die(true,"pthread_join()");
}

