
/*
 *    ChunkSync - create space-efficient incremental (remote) backups of large files
 *    Copyright (C) 2009  Florian Zumbiehl <florz@florz.de>
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 *
 *    In addition, as a special exception, the author of this
 *    program gives permission to link the code of its 
 *    release with the OpenSSL project's "OpenSSL" library (or
 *    with modified versions of it that use the same license as
 *    the "OpenSSL" library), and distribute the linked
 *    executables. You must obey the GNU General Public 
 *    License in all respects for all of the code used other 
 *    than "OpenSSL".  If you modify this file, you may extend
 *    this exception to your version of the file, but you are
 *    not obligated to do so.  If you do not wish to do so,
 *    delete this exception statement from your version.
 */

#define _ATFILE_SOURCE

#include <stdbool.h>
#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/uio.h>

#include <openssl/evp.h>

#include "utils.h"
#include "defs.h"
#include "io.h"
#include "backend.h"
#include "err.h"
#include "frontend.h"
#include "receiver.h"

static pthread_mutex_t send_lock=PTHREAD_MUTEX_INITIALIZER;
static size_t linkdirs_cnt;
static int linkdirs_fds[LINKDIRS_MAX];
static char *linkdirs_names[LINKDIRS_MAX];

static void chunk_name_init(char *buf){
	memcpy(buf+24,"digest",7);
}

static void chunk_name_gen(char *buf,off_t num){
	sprintf(buf+7,"%016llx",num);
	for(int x=0;x<7;x++){
		memmove(buf+x*3,buf+7+x*2,2);
		buf[x*3+2]='/';
	}
}

static void chunk_name_digest(char *buf){
	buf[23]='.';
}

static void chunk_name_undigest(char *buf){
	buf[23]=0;
}

static void ensuredirexists(int rootdir_fd,off_t chunk_num,char *chunk_name){
	int x;

	for(x=0;x<8&&!(chunk_num&0xff);chunk_num>>=8,x++);
	for(x=max(7-x,0);x<7;x++){
		chunk_name[x*3+2]=0;
		if(mkdirat(rootdir_fd,chunk_name,0777))die(true,"mkdirat(<destination>/%s)",chunk_name);
		chunk_name[x*3+2]='/';
	}
}

static void digest_binary_to_text(char *text,const uint8_t *binary){
	for(int x=0;x<DIGEST_SALT_SIZE+DIGEST_SIZE;x++)
		sprintf(text+x*2,"%02x",binary[x]);
}

static void digest_text_to_binary(uint8_t *binary,const char *text,const char *linkdir_name,const char *chunk_name){
	uint8_t tmp;

	for(size_t x=0;x<(DIGEST_SALT_SIZE+DIGEST_SIZE)*2;x++){
		tmp=text[x];
		if(tmp>='0'&&tmp<='9'){
			tmp-='0';
		}else if(tmp>='a'&&tmp<='f'){
			tmp-='a'-10;
		}else{
			die(false,"%s/%s corrupted",linkdir_name,chunk_name);
		}
		binary[x/2]=(binary[x/2]<<4)|tmp;
	}
}

static void *linkdest_tx(void *ignored){
	bool linkdirs_eof[LINKDIRS_MAX];
	size_t linkdirs_eof_cnt,linkdir_num;
	uint64_t chunk_num;
	off_t prev_chunk_size;
	char chunk_name[23+7+1],chunk_digest_text[(DIGEST_SALT_SIZE+DIGEST_SIZE)*2+1];
	struct stat st;
	unsigned char chunk_digest[DIGEST_SALT_SIZE+DIGEST_SIZE],prev_chunk_digest[DIGEST_SALT_SIZE+DIGEST_SIZE];
	int r;
	uint8_t transfer;

	memset(linkdirs_eof,0,sizeof(linkdirs_eof));
	prev_chunk_size=1;
	memset(prev_chunk_digest,0,sizeof(prev_chunk_digest));
	chunk_name_init(chunk_name);
	for(linkdirs_eof_cnt=chunk_num=0;linkdirs_eof_cnt<linkdirs_cnt;chunk_num++){
		chunk_name_gen(chunk_name,chunk_num);
		for(linkdir_num=0;linkdir_num<linkdirs_cnt;linkdir_num++)if(!linkdirs_eof[linkdir_num]){
			chunk_name_undigest(chunk_name);
			errno=ENOENT;
			if(chunk_num==ANYSINT_MAX(off_t)||fstatat(linkdirs_fds[linkdir_num],chunk_name,&st,0)){
				if(errno!=ENOENT)die(true,"fstatat(%s/%s)",linkdirs_names[linkdir_num],chunk_name);
				pthread_mutex_lock(&send_lock);
				xwrite_u8(&stdio_fh,0<<3);
				pthread_mutex_unlock(&send_lock);
				linkdirs_eof[linkdir_num]=true;
				linkdirs_eof_cnt++;
			}else{
				if(st.st_size<1)die(false,"%s/%s corrupted",linkdirs_names[linkdir_num],chunk_name);
				chunk_name_digest(chunk_name);
				if((r=readlinkat(linkdirs_fds[linkdir_num],chunk_name,chunk_digest_text,sizeof(chunk_digest_text)))!=(DIGEST_SALT_SIZE+DIGEST_SIZE)*2)die(r<0,"readlinkat(%s/%s)",linkdirs_names[linkdir_num],chunk_name);
				digest_text_to_binary(chunk_digest,chunk_digest_text,linkdirs_names[linkdir_num],chunk_name);
				transfer=(st.st_size!=prev_chunk_size)
					|(memcmp(chunk_digest,prev_chunk_digest,DIGEST_SALT_SIZE)?2:0)
					|(memcmp(chunk_digest+DIGEST_SALT_SIZE,prev_chunk_digest+DIGEST_SALT_SIZE,DIGEST_SIZE)?4:0);
				pthread_mutex_lock(&send_lock);
				xwrite_u8(&stdio_fh,(1<<3)|transfer);
				if(transfer&1)xwrite_u64(&stdio_fh,st.st_size);
				if(transfer&2)xwrite(&stdio_fh,chunk_digest,DIGEST_SALT_SIZE);
				if(transfer&4)xwrite(&stdio_fh,chunk_digest+DIGEST_SALT_SIZE,DIGEST_SIZE);
				pthread_mutex_unlock(&send_lock);
				prev_chunk_size=st.st_size;
				memcpy(prev_chunk_digest,chunk_digest,DIGEST_SALT_SIZE+DIGEST_SIZE);
			}
		}
	}
	pthread_exit(NULL);
}

void receiver_main(const char *destdir_name){
	char chunk_name[23+7+1],chunk_digest_text[(DIGEST_SALT_SIZE+DIGEST_SIZE)*2+1];
	uint64_t chunk_size;
	off_t chunk_num;
	int destdir_fd,chunk_fd;
	struct xfh chunk_fh;
	pthread_t thr;
	unsigned char chunk_digest[DIGEST_SALT_SIZE+DIGEST_SIZE];
	size_t linkdir_num;

	xwrite_u8(&stdio_fh,2);
	if((linkdirs_cnt=xread_u8(&stdio_fh))>LINKDIRS_MAX)proterr();
	for(linkdir_num=0;linkdir_num<linkdirs_cnt;linkdir_num++){
		linkdirs_names[linkdir_num]=xread_string(&stdio_fh);
		if((linkdirs_fds[linkdir_num]=open(linkdirs_names[linkdir_num],O_RDONLY))<0)die(true,"open(%s)",linkdirs_names[linkdir_num]);
	}
	if(xread_u8(&stdio_fh))proterr();
	if(mkdir(destdir_name,0777))die(true,"mkdir(destination)");
	if((destdir_fd=open(destdir_name,O_RDONLY))<0)die(true,"open(destination)");
	if((errno=pthread_create(&thr,NULL,linkdest_tx,NULL)))die(true,"pthread_create()");
	chunk_name_init(chunk_name);
	for(chunk_num=0;true;chunk_num++){
		chunk_name_gen(chunk_name,chunk_num);
		switch((linkdir_num=xread_u8(&stdio_fh))>>6){
			case 0:
				linkdir_num&=0x3f;
				if(linkdir_num>=linkdirs_cnt)proterr();
				if(chunk_num==ANYSINT_MAX(off_t))proterr();
				ensuredirexists(destdir_fd,chunk_num,chunk_name);
				if(linkat(linkdirs_fds[linkdir_num],chunk_name,destdir_fd,chunk_name,0))die(true,"linkat(%s/%s)",linkdirs_names[linkdir_num],chunk_name);
				chunk_name_digest(chunk_name);
				if(linkat(linkdirs_fds[linkdir_num],chunk_name,destdir_fd,chunk_name,0))die(true,"linkat(%s/%s)",linkdirs_names[linkdir_num],chunk_name);
				break;
			case 1:
				if(linkdir_num&0x3f)proterr();
				if(chunk_num==ANYSINT_MAX(off_t))proterr();
				chunk_size=xread_u64(&stdio_fh);
				if(chunk_size<1||chunk_size>ANYSINT_MAX(off_t))proterr();
				xread(&stdio_fh,chunk_digest,DIGEST_SALT_SIZE+DIGEST_SIZE);
				digest_binary_to_text(chunk_digest_text,chunk_digest);
				ensuredirexists(destdir_fd,chunk_num,chunk_name);
				if((chunk_fd=openat(destdir_fd,chunk_name,O_WRONLY|O_CREAT|O_EXCL,0666))<0)die(true,"openat(<destination>/%s)",chunk_name);
				xfh_init(&chunk_fh,-1,chunk_fd,NULL,"chunk");
				xcopy(&chunk_fh,&stdio_fh,chunk_size);
				if(close(chunk_fd))die(true,"close(<destination>/%s)",chunk_name);
				chunk_name_digest(chunk_name);
				if(symlinkat(chunk_digest_text,destdir_fd,chunk_name))die(true,"symlinkat(<destination>/%s)",chunk_name);
				break;
			case 2:
				if(linkdir_num&0x3f)proterr();
				if(close(destdir_fd))die(true,"close(destination)");
				pthread_mutex_lock(&send_lock);
				xwrite_u8(&stdio_fh,2<<3);
				return;
			default:
				proterr();
		}
	}
}

void receiver_forward(){
	uint8_t tmp;

	while(true){
		switch((tmp=xread_u8(&receiver_fh))>>3){
			case 0:
				xwrite_u8(&sender_fh,0<<3);
				break;
			case 1:
				xwrite_u8(&sender_fh,tmp);
				if(tmp&1)xwrite_u64(&sender_fh,xread_u64(&receiver_fh));
				if(tmp&2)xcopy(&sender_fh,&receiver_fh,DIGEST_SALT_SIZE);
				if(tmp&4)xcopy(&sender_fh,&receiver_fh,DIGEST_SIZE);
				break;
			case 2:
				xwrite_u8(&sender_fh,2<<3);
				return;
			default:
				proterr();
		}
	}
}

